/* This file is part of PolyCalc.

   PolyCalc is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   any later version.

   PolyCalc is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with PolyCalc.  If not, see <http://www.gnu.org/licenses/>.

   For more information about PolyCalc see project home
   <https://launchpad.net/polycalc>.

   Copyright 2012 PolyCalc development team */

#include "mainwindow.h"
#include "ui_mainwindow.h"
#include <algorithm>
#include <QPushButton>
#include <QLabel>
#include <QLineEdit>
#include <QMessageBox>
#include <QDebug>
#include <QFileDialog>
#include <QFileInfo>
#include <QIcon>
#include <QtXml>
#include <QErrorMessage>
#include <QtGlobal>

const QString version = "0.1.0";
const QString prjUrl  = "launchpad.net/polycalc";
const QString helpUrl = "http://illusions.hu/index.php?task=100&statpage=252";
#ifdef QT_DEBUG
  const QString appdir  = "../polycalc/layouts";
#else
  #ifdef Q_OS_WIN32
    const QString appdir  = ".";
  #else
    const QString appdir  = "/usr/share/polycalc";
  #endif
#endif

class ButtonAction : public QObject {
  public:
    static int buttoncnt;
    int id;
    QString insert;
    QString set;
    QString append;
    QString prepend;
    bool send;
    ButtonAction(QString insert,
                 QString set,
                 QString append,
                 QString prepend,
                 bool send)
               : insert(insert),
                 set(set),
                 append(append),
                 prepend(prepend),
                 send(send)
    {
      id = buttoncnt;
      buttoncnt++;
    }
};

int ButtonAction::buttoncnt=0;

MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow)
{
    menuSignalMapper   = new QSignalMapper(this);
    buttonSignalMapper = new QSignalMapper(this);
    scriptEngine       = new QScriptEngine();
    externalEngine     = new QProcess(this);
    stringPipeEngine   = new QScriptEngine();
    engineType         = eEngTypeQtScript;
    settings           = new QSettings("PolyCalc development team", "PolyCalc");
    statusbarScript    = "error";
    ui->setupUi(this);
    ui->dockWidget->hide();
    QWidget::setWindowIcon(QIcon(appdir + "/icon256.png"));
    PopulateMenu();
    loadLayout(settings->value("layout", "basic").toString());
    setStatusBarText("started");
    ui->lineEdit->insertItem(0,"");
    ui->lineEdit->setCurrentIndex(0);
    ui->lineEdit->lineEdit()->setAlignment(Qt::AlignRight);
    connect(ui->lineEdit->lineEdit(), SIGNAL(returnPressed(void)),
            this, SLOT(on_lineEdit_returnPressed(void)));
    resize(settings->value("window_width", 0).toInt(),
           settings->value("window_height", 0).toInt());
}

void MainWindow::PopulateMenu()
{
    QDir dir(appdir);
    QStringList list = dir.entryList(QStringList("*.cal"));
    QStringListIterator iterator(list);
    while (iterator.hasNext()) {
        QString name = iterator.next();
        name.chop(4);
        QAction *action = new QAction(name, this);
        ui->menuOptions->addAction(action);
        menuSignalMapper->setMapping(action, name);
        connect(action, SIGNAL(triggered()),
                menuSignalMapper, SLOT (map()));
    }
    connect(menuSignalMapper, SIGNAL(mapped(const QString &)),
            this, SLOT(loadLayout(const QString &)));
}

void MainWindow::PopulateButtons(QString filename)
{
    static QGridLayout *rootGridWidget = NULL;
    QDomDocument doc("buttontable");
    QFile file(filename);
    QString errorMsg;
    int errorLine;
    int errorColumn;

    if (!file.open(QIODevice::ReadOnly)) {
        QErrorMessage errorMessage;
        errorMessage.showMessage(QString("Can not open file ") + filename);
        errorMessage.exec();
    }

    if (!doc.setContent(&file, true, &errorMsg, &errorLine, &errorColumn)) {
        file.close();
        QErrorMessage errorMessage;
        errorMessage.showMessage(QString("Error while processing file ") + filename
                                 + QString(" at line ") + QString::number(errorLine)
                                 + QString(", column ") + QString::number(errorColumn)
                                 + QString(": ") + errorMsg);
        errorMessage.exec();
    }
    file.close();

    if (rootGridWidget != NULL) {
        delete buttonSignalMapper;
        buttonSignalMapper = new QSignalMapper(this);
        deleteGrid(rootGridWidget);
        ui->verticalLayout->removeItem(rootGridWidget);
        delete rootGridWidget;
        rootGridWidget=NULL;
    }
    QDomElement polycalc = doc.documentElement();    
    for (QDomNode rootNode = polycalc.firstChild();
         !rootNode.isNull();
         rootNode = rootNode.nextSibling())
    {
        if (rootNode.toElement().tagName() == "buttontable")
        {
            QDomNode rootgrid = rootNode.firstChild();
            if(rootgrid.toElement().tagName() == "grid") {
                rootGridWidget = XMLGridNodeToQGridLayout(&rootgrid,
                                                          rootgrid.toElement().attribute("style",NULL));
                ui->verticalLayout->addLayout(rootGridWidget);
            }
        }
        if (rootNode.toElement().tagName() == "hooks")
        {
            for (QDomNode hook = rootNode.firstChild();
                 !hook.isNull();
                 hook = hook.nextSibling())
            {
                if (hook.toElement().tagName() == "init")
                {
                    QScriptValue res = stringPipeEngine->evaluate(rootNode.toElement().text());
                    if (res.isError()) {
                        // QErrorMessage errorMessage;
                        // errorMessage.showMessage(res.toString());
                        // errorMessage.exec();
                    }
                } else {
                    QScriptValue res =
                      stringPipeEngine->evaluate("function global"+hook.toElement().tagName()+"() "+
                                                 "{ "+rootNode.toElement().text()+" }");
                    if (res.isError()) {
                       // QErrorMessage errorMessage;
                       // errorMessage.showMessage(res.toString());
                       // errorMessage.exec();
                    }
                }
            }
        }
        if (rootNode.toElement().tagName() == "statusbar")
        {
            statusbarScript = rootNode.toElement().attribute("text");
        }
        if (rootNode.toElement().tagName() == "engine")
        {
            if (rootNode.toElement().attribute("type") == "external")
            {
                engineType = eEngTypeExternal;
                QString exe = rootNode.toElement().attribute("command");
                externalEngine->start(exe, QIODevice::ReadWrite);
                if (!externalEngine->waitForStarted()) {
                    QErrorMessage errorMessage;
                    errorMessage.showMessage("Can not find executable '"+exe+"'.");
                    errorMessage.exec();
                }
                connect(externalEngine, SIGNAL(readyReadStandardOutput()),this, SLOT(on_extEngStdOut()));
                connect(externalEngine, SIGNAL(readyReadStandardError ()),this, SLOT(on_extEngStdErr()));
            }
            else
            if (rootNode.toElement().attribute("type") == "QtScript")
            {
                engineType = eEngTypeQtScript;
                for (QDomNode scriptElement = rootNode.firstChild();
                     !scriptElement.isNull();
                     scriptElement = scriptElement.nextSibling())
                {
                     if (scriptElement.toElement().tagName() == "init")
                     {
                        // todo: chech engine type
                        scriptEngine->evaluate(scriptElement.toElement().text());
                     }
                }
            }
            else
            {
                // todo: not supported engine
            }
        }

    }
    connect(buttonSignalMapper, SIGNAL(mapped(QObject *)),
            this, SLOT(on_anyButtonClicked(QObject *)));
}

void MainWindow::on_extEngStdOut()
{
}

void MainWindow::on_extEngStdErr()
{
    setStatusBarText(QString(externalEngine->readAllStandardError()));
}

void MainWindow::deleteGrid(QLayout *grid)
{
    QLayoutItem* item;
    while ( (item = grid->takeAt(0)) != NULL )
    {
        if (item->layout() != 0) deleteGrid(item->layout());
        delete item->widget();
        delete item;
    }
}

QGridLayout* MainWindow::XMLGridNodeToQGridLayout(QDomNode *grid, QString parentStyle)
{
    int i = 0, j = 0;
    QGridLayout *gridWidget = new QGridLayout();
    gridWidget->setMargin(0);
    gridWidget->setSpacing(grid->toElement().attribute("spacing",0).toInt());
    for (QDomNode row = grid->firstChild();
         !row.isNull();
         row = row.nextSibling(), j++)
    {
        i = 0;
        QString rowStyle = row.toElement().attribute("style",NULL);
        int rowStretch = row.toElement().attribute("stretch", "1").toInt();
        gridWidget->setRowStretch(j,rowStretch);
        for (QDomNode cell = row.firstChild();
             !cell.isNull();
             cell = cell.nextSibling(), i++)
        {
            int colStretch = cell.toElement().attribute("stretch", "1").toInt();
            gridWidget->setColumnStretch(i,colStretch);
            QString cellStyle = cell.toElement().attribute("style",NULL);
            QString style = (cellStyle != NULL) ? cellStyle :
                            (rowStyle != NULL)  ? rowStyle  :
                                                  parentStyle;
            if(cell.toElement().tagName() == "button")
            {
                ButtonAction *buttonAction;
                QString caption = cell.toElement().attribute("caption");
                QPushButton *button = new QPushButton(caption, this);
                button->setGeometry(0,0,45,25);
                button->setMinimumSize(45,25);
                button->setFocusPolicy(Qt::NoFocus);
                button->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
                button->setStyleSheet(style);
                gridWidget->addWidget(button, j, i);
                buttonAction = new ButtonAction(cell.toElement().attribute("insert",  NULL),
                                                cell.toElement().attribute("set",     NULL),
                                                cell.toElement().attribute("append",  NULL),
                                                cell.toElement().attribute("prepend", NULL),
                                                cell.toElement().attribute("send") == "true");
                QString hooks[] = {"presend","postsend","preshow"};
                for (int i=0; i < 3; i++)
                {
                    if (cell.toElement().attribute(hooks[i],NULL) != NULL)
                    {
                        stringPipeEngine->evaluate("function button"+
                                                   QString::number(buttonAction->id)+hooks[i]+"() "+
                                                    "{ "+cell.toElement().attribute(hooks[i])+" }");
                    }
                }
                buttonSignalMapper->setMapping(button, buttonAction);
                connect(button, SIGNAL(clicked()),
                        buttonSignalMapper, SLOT (map()));
            }
            if(cell.toElement().tagName() == "label")
            {
                QLabel *label = new QLabel(cell.toElement().attribute("caption"), this);
                QString name = cell.toElement().attribute("name");
                if (name != "") {
                    label->setObjectName(name);
                    labels.push_back(label);
                }
                label->setGeometry(0,0,40,15);
                label->setMinimumSize(0,15);
                label->setFocusPolicy(Qt::NoFocus);
                label->setAlignment(Qt::AlignHCenter|Qt::AlignBottom);
                label->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
                label->setStyleSheet(style);
                gridWidget->addWidget(label, j, i);
            }
            if(cell.toElement().tagName() == "grid")
            {
                QGridLayout *childGridWidget = XMLGridNodeToQGridLayout(&cell, style);
                gridWidget->addLayout(childGridWidget, j, i);
            }
        }
    }
    return(gridWidget);
}

MainWindow::~MainWindow()
{
    switch(engineType) {
      case eEngTypeExternal: {
          externalEngine->closeWriteChannel();
          externalEngine->closeReadChannel(QProcess::StandardOutput);
          externalEngine->closeReadChannel(QProcess::StandardError);
          externalEngine->close();
          break;
      }
    }
    delete ui;
}

void MainWindow::updateLabels(void)
{
    for (int i = 0; i < labels.size(); ++i) {
        QString name = labels.at(i)->objectName();
        QString caption = stringPipeEngine->evaluate(name+";").toString();
        labels.at(i)->setText(caption);
    }
}

void MainWindow::resizeEvent(QResizeEvent *event)
{
    settings->setValue("window_width", width());
    settings->setValue("window_height", height());
}

void MainWindow::loadLayout(const QString &name)
{
    settings->setValue("layout",name);
    this->setWindowTitle("PolyCalc - "+name);
    PopulateButtons(appdir + "/" + name + QString(".cal"));
    resize(0,0);
}

void MainWindow::on_anyButtonClicked(QObject * obj)
{
    ButtonAction *action = (ButtonAction*)obj; 
    qDebug() << action->id;
    if (!action->insert.isNull()) {
        ui->lineEdit->lineEdit()->insert(action->insert);
    }
    if (!action->set.isNull()) {
        ui->lineEdit->lineEdit()->setText(action->set);
    }
    if (!action->append.isNull()) {
        ui->lineEdit->lineEdit()->setText(
                    ui->lineEdit->lineEdit()->text().append(action->append));
    }
    if (!action->prepend.isNull()) {
        ui->lineEdit->lineEdit()->setText(
                    ui->lineEdit->lineEdit()->text().prepend(action->prepend));
    }
    QString display = QString(ui->lineEdit->currentText());
    stringPipeEngine->evaluate("cursorpos = \""
                               +QString::number(ui->lineEdit->lineEdit()->cursorPosition())+"\";");
    stringPipeEngine->evaluate("beforeselection  = \""+
                               ui->lineEdit->lineEdit()->text().left(
                                   std::min(ui->lineEdit->lineEdit()->cursorPosition(),
                                            ui->lineEdit->lineEdit()->selectionStart()))+"\";");
    stringPipeEngine->evaluate("beforecursor  = \""+
                               ui->lineEdit->lineEdit()->text().left(
                                   ui->lineEdit->lineEdit()->cursorPosition())+"\";");
    stringPipeEngine->evaluate("selected = \""+ui->lineEdit->lineEdit()->selectedText()+"\";");
    stringPipeEngine->evaluate("aftercursor  = \""+
                               ui->lineEdit->lineEdit()->text().right(
                                   ui->lineEdit->lineEdit()->text().size() -
                                   ui->lineEdit->lineEdit()->cursorPosition())+"\";");
    stringPipeEngine->evaluate("afterselection  = \""+
                               ui->lineEdit->lineEdit()->text().right(
                                   ui->lineEdit->lineEdit()->text().size() -
                                   std::max(ui->lineEdit->lineEdit()->cursorPosition(),
                                            ui->lineEdit->lineEdit()->selectionStart()))+"\";");

    QString error("no error");
    if (action->send) {
        ui->plainTextEdit->appendPlainText("> "+ui->lineEdit->currentText().trimmed());
        QString reply("");
        display = QString(ui->lineEdit->currentText());
        stringPipeEngine->evaluate("display = \""+display+"\";");
        stringPipeEngine->evaluate("command = display;");
        stringPipeEngine->evaluate("globalpresend();");
        stringPipeEngine->evaluate("button"+QString::number(action->id)+"presend();");
        QString command = stringPipeEngine->evaluate("command;").toString();
        qDebug() << command;
        switch(engineType) {
          case eEngTypeQtScript: {
              QScriptValue result = scriptEngine->evaluate(command);
              if (result.isError()) {
                  error = result.toString();
                  ui->plainTextEdit->appendPlainText(error);
              } else {
                  reply = result.toString();
              }
              break;
          }
          case eEngTypeExternal: {
              command += "\n";
              if (externalEngine->write(command.toAscii()) == -1) {
                  error = "Can not write standard input of external program: "+
                                              externalEngine->errorString();
                  ui->plainTextEdit->appendPlainText(error);
              }
              externalEngine->waitForBytesWritten(1000);
              externalEngine->waitForReadyRead(1000);
              reply = QString(externalEngine->readAllStandardOutput()).trimmed();
              break;
          }
        }
        qDebug() << reply;
        stringPipeEngine->evaluate("reply = \""+reply+"\";");
        stringPipeEngine->evaluate("globalpostsend();");
        stringPipeEngine->evaluate("button"+QString::number(action->id)+"postsend();");
        display = stringPipeEngine->evaluate("reply;").toString();
        qDebug() << display;
    }
    stringPipeEngine->evaluate("display = \""+display+"\";");
    stringPipeEngine->evaluate("globalpreshow();");
    stringPipeEngine->evaluate("button"+QString::number(action->id)+"preshow();");
    display = stringPipeEngine->evaluate("display;").toString();
    if (action->send) {
        ui->plainTextEdit->appendPlainText(display);
        ui->lineEdit->insertItem(0,display );
        ui->lineEdit->setCurrentIndex(0);
    } else {
        ui->lineEdit->lineEdit()->setText(display);
    }
    setStatusBarText(error);
    updateLabels();
}

void MainWindow::setStatusBarText(QString error)
{
    stringPipeEngine->evaluate("error = \""+error+"\";");
    stringPipeEngine->evaluate("statusbar = "+statusbarScript+";");
    QString statusbar = stringPipeEngine->evaluate("statusbar;").toString();
    ui->statusBar->showMessage(statusbar);
}

void MainWindow::on_lineEdit_returnPressed()
{
    static ButtonAction send(NULL, NULL, NULL, NULL, true);
    on_anyButtonClicked(&send);
}

void MainWindow::on_actionQuit_triggered()
{
    this->close();
}

void MainWindow::on_actionAbout_triggered()
{
    QMessageBox msgBox;
    msgBox.setTextFormat(Qt::RichText);
    msgBox.about(this,
                 "About",
                 "<b>PolyCalc v" + version + "</b><br />" +
                 "Configurable UI Calculator<br />" +
                 QString::fromUtf8("© 2012 PolyCalc Development Team<br />") +
                 "Distributed under GNU GPL v3<br />" +
                 "<a href='http://"+prjUrl+"'>"+prjUrl+"</a>");
}

void MainWindow::on_actionHelp_triggered()
{
    QMessageBox msgBox;
    msgBox.setTextFormat(Qt::RichText);
    msgBox.setText("For help please visit PolyCalc Help page:<br />"
                   "<a href='http://"+helpUrl+"'>"+helpUrl+"</a>");
    msgBox.exec();
}

void MainWindow::on_actionCommand_Log_triggered()
{
    ui->dockWidget->setVisible(ui->actionCommand_Log->isChecked());
}

